/*
 * This file is part of PropertiesOrderedMap library.
 *
 * Copyright 2018 Arduino AG (http://www.arduino.cc/)
 *
 * PropertiesOrderedMap library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * As a special exception, you may use this file as part of a free software
 * library without restriction.  Specifically, if other files instantiate
 * templates or use macros or inline functions from this file, or you compile
 * this file and link it with other files to produce an executable, this
 * file does not by itself cause the resulting executable to be covered by
 * the GNU General Public License.  This exception does not however
 * invalidate any other reasons why the executable file might be covered by
 * the GNU General Public License.
 */

package properties

import (
	"encoding/json"
	"testing"

	"github.com/stretchr/testify/require"
)

func TestJson(t *testing.T) {
	testmap := NewMap()
	testmap.Set("a", "1")
	testmap.Set("b", "2")
	testmap.Set(`"e"`, "3")
	testmap.Set("d", "4")
	testmap.Set(`\"c"\`, "5")

	d, err := json.Marshal(testmap)
	require.NoError(t, err)
	require.Equal(t, `{"\"e\"":"3","\\\"c\"\\":"5","a":"1","b":"2","d":"4"}`, string(d))

	var decodedmap *Map
	err = json.Unmarshal(d, &decodedmap)
	require.NoError(t, err)
	require.Equal(t, 5, decodedmap.Size())
	require.Equal(t, "1", decodedmap.Get("a"))
	require.Equal(t, "2", decodedmap.Get("b"))
	require.Equal(t, "3", decodedmap.Get(`"e"`))
	require.Equal(t, "4", decodedmap.Get("d"))
	require.Equal(t, "5", decodedmap.Get(`\"c"\`))
}
